<?php

namespace Drupal\interactive_queue_processing;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\QueueWorkerManagerInterface;
use Drupal\Core\Queue\RequeueException;
use Drupal\Core\Queue\SuspendQueueException;

/**
 * Various queue management functionality.
 */
class QueueManager {

  /**
   * All queue definitions.
   *
   * @var array|null
   */
  protected $queues = NULL;

  /**
   * The config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The queue factory service.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * The queue worker plugin manager service.
   *
   * @var \Drupal\Core\Queue\QueueWorkerManagerInterface
   */
  protected $queueWorkerManager;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * QueueManager constructor.
   *
   * @param \Drupal\Core\Queue\QueueFactory $queueFactory
   *   The queue factory service.
   * @param \Drupal\Core\Queue\QueueWorkerManagerInterface $queueWorkerManager
   *   The queue worker plugin manager service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory service.
   */
  public function __construct(QueueFactory $queueFactory, QueueWorkerManagerInterface $queueWorkerManager, TimeInterface $time, EntityTypeManagerInterface $entityTypeManager, ConfigFactoryInterface $configFactory) {
    $this->queueFactory = $queueFactory;
    $this->queueWorkerManager = $queueWorkerManager;
    $this->time = $time;
    $this->entityTypeManager = $entityTypeManager;
    $this->configFactory = $configFactory;
  }

  /**
   * Gets the queue definitions.
   *
   * @return array
   *   Queue definitions.
   */
  public function getQueueDefinitions() {
    if ($this->queues !== NULL) {
      return $this->queues;
    }

    $config = $this->configFactory->get('interactive_queue_processing.settings');
    $queues = $config->get('queues');

    $defaults = [
      'status' => $config->get('default_queue_status'),
      'priority' => $config->get('default_queue_priority'),
    ];

    foreach ($this->getCoreQueues() as $queue) {
      $queues = $this->setQueueDefinition($queues, $queue['id'], 'core', $defaults + ['queue' => $queue]);
    }

    foreach ($this->getAdvancedQueues() as $queue) {
      $queues = $this->setQueueDefinition($queues, $queue->id(), 'advancedqueue_queue', $defaults + ['queue' => $queue]);
    }

    usort($queues, '\Drupal\interactive_queue_processing\QueueManager::sortQueues');
    $this->queues = $queues;
    return $this->queues;
  }

  /**
   * Set queue definition into queues array.
   *
   * @param array $queues
   *   The queues array.
   * @param string $name
   *   The queue name/ID.
   * @param string $type
   *   The queue type.
   * @param mixed $queue
   *   The queue definition.
   *
   * @return array
   *   The updated queues array.
   */
  protected function setQueueDefinition(array $queues, $name, $type, $queue) {
    // If queue already exists in queues array, tack on the definition and leave
    // the rest as-is.
    foreach ($queues as $index => $item) {
      $queues[$index]['queue'] = $queue['queue'];
      if ($item['name'] == $name && $item['type'] == $type) {
        return $queues;
      }
    }

    // Did not already exist, add definition to queues array.
    $queue['name'] = $name;
    $queue['type'] = $type;
    $queues[] = $queue;
    return $queues;
  }

  /**
   * Sort callback for ordering of queues.
   *
   * @param array $a
   *   Queue definition.
   * @param array $b
   *   Queue definition.
   *
   * @return int
   *   Sort indicator.
   */
  protected static function sortQueues(array $a, array $b) {
    // Primary sort, priority (highest first).
    if ($a['priority'] < $b['priority']) {
      return 1;
    }
    elseif ($a['priority'] > $b['priority']) {
      return -1;
    }

    // Secondary sort, name (alpha).
    if ($a['name'] > $b['name']) {
      return 1;
    }
    elseif ($a['name'] < $b['name']) {
      return -1;
    }

    // Tertiary sort, queue type (alpha).
    if ($a['type'] > $b['type']) {
      return 1;
    }
    elseif ($a['type'] < $b['type']) {
      return -1;
    }

    return 0;
  }

  /**
   * Get the core queues.
   *
   * @return array
   *   The core queues.
   */
  protected function getCoreQueues() {
    return array_filter($this->queueWorkerManager->getDefinitions(), function ($info) {
      return isset($info['cron']);
    });
  }

  /**
   * Get the advanced queues.
   *
   * @return array|\Drupal\advancedqueue\Entity\Queue[]
   *   Advanced queues.
   */
  protected function getAdvancedQueues() {
    if (!$this->entityTypeManager->getDefinition('advancedqueue_queue', FALSE)) {
      return [];
    }

    /** @var \Drupal\advancedqueue\Entity\Queue[] $advancedQueues */
    $advancedQueues = $this->entityTypeManager->getStorage('advancedqueue_queue')
      ->loadMultiple();
    return array_filter($advancedQueues, function ($queue) {
      return $queue->getProcessor() == 'cron';
    });
  }

  /**
   * Processes cron queues.
   *
   * @see \Drupal\Core\Cron::processQueues
   */
  public function processQueues() {
    $config = $this->configFactory->get('interactive_queue_processing.settings');

    // We'll only process for ~the configured amount of time.
    $start = $this->time->getCurrentMicroTime();
    $end = $start + $config->get('processing_time');

    while ($this->time->getCurrentMicroTime() < $end) {
      foreach ($this->getQueueDefinitions() as $definition) {
        // Only enabled queues, already ordered by priority.
        if (empty($definition['status'])) {
          return;
        }

        switch ($definition['type']) {
          case 'advancedqueue_queue':
            $this->processAdvancedqueueQueue($definition, $end);
            break;

          case 'core':
            $this->processCoreQueue($definition, $end);
            break;
        }
      }
    }
  }

  /**
   * Process an advancedqueue queue.
   *
   * @param array $queue_definition
   *   The queue definition.
   * @param int $end
   *   The time at which the processing needs to end.
   */
  protected function processAdvancedqueueQueue(array $queue_definition, $end) {
    // Ensure entity type exists (module installed).
    if (!$this->entityTypeManager->getDefinition('advancedqueue_queue', FALSE)) {
      return;
    }

    /** @var \Drupal\advancedqueue\Entity\Queue $queue */
    $queue = $this->entityTypeManager->getStorage('advancedqueue_queue')
      ->load($queue_definition['name']);

    // Ensure we have a queue.
    if (!$queue) {
      return;
    }

    // @todo dependency injection, possible when service may not exist?
    /** @var \Drupal\advancedqueue\Processor $processor */
    $processor = \Drupal::service('advancedqueue.processor');

    // Loop until we've run out of time or there are no jobs to claim.
    while ($this->time->getCurrentMicroTime() < $end && ($job = $queue->getBackend()->claimJob())) {
      $processor->processJob($job, $queue);
    }
  }

  /**
   * Process a core queue.
   *
   * @param array $queue_definition
   *   The queue definition.
   * @param int $end
   *   The time at which the processing needs to end.
   */
  protected function processCoreQueue(array $queue_definition, $end) {
    $this->queueFactory->get($queue_definition['name'])->createQueue();
    $queue_worker = $this->queueWorkerManager->createInstance($queue_definition['name']);
    $queue = $this->queueFactory->get($queue_definition['name']);
    $lease_time = isset($queue_definition['queue']['cron']['time']) ?: NULL;
    while ($this->time->getCurrentMicroTime() < $end && ($item = $queue->claimItem($lease_time))) {
      try {
        $queue_worker->processItem($item->data);
        $queue->deleteItem($item);
      }
      catch (RequeueException $e) {
        // The worker requested the task be immediately requeued.
        $queue->releaseItem($item);
      }
      catch (SuspendQueueException $e) {
        // If the worker indicates there is a problem with the whole queue,
        // release the item and skip to the next queue.
        $queue->releaseItem($item);

        watchdog_exception('interactive_queue_processing', $e);

        // Skip to the next queue.
        return;
      }
      catch (\Exception $e) {
        // In case of any other kind of exception, log it and leave the item
        // in the queue to be processed again later.
        watchdog_exception('interactive_queue_processing', $e);
      }
    }
  }

  /**
   * Counts items in queues.
   *
   * @param bool $sum
   *   Return the sum of all remaining items. Optional, defaults to TRUE.
   *
   * @return array
   *   Number of items in all queues remaining to be processed, or an array
   *   of queue information including counts.
   */
  public function queueCounts($sum = TRUE) {
    $count = [];

    foreach ($this->getQueueDefinitions() as $info) {
      if (empty($info['status'])) {
        continue;
      }

      $row = [
        'name' => $info['name'],
        'type' => $info['type'],
        'count' => 0,
      ];

      switch ($row['type']) {
        case 'advancedqueue_queue':
          if ($this->entityTypeManager->getDefinition('advancedqueue_queue', FALSE)) {
            /** @var \Drupal\advancedqueue\Entity\Queue $queue */
            if ($queue = $this->entityTypeManager->getStorage('advancedqueue_queue')->load($row['name'])) {
              $row['count'] = $queue->getBackend()->countJobs()['queued'];
            }
          }
          break;

        case 'core':
          $queue = $this->queueFactory->get($info['name']);
          $row['count'] = $queue->numberOfItems();
          break;
      }

      $count[] = $row;
    }


    if ($sum) {
      $count = array_map(function ($row) {
        return $row['count'];
      }, $count);
      $count = array_sum($count);
    }

    return $count;
  }

}
